"""
  Module to conduct evaluation on a given record linkage clustering result.

  The evaluation method we have used here is the one proposed in our journal 
  paper 'Evaluation measure for group-based record linkage'.
  
  This evaluation method assessed how well the record have been assigned to
  the predicted clusters, with respect to how they appear in the ground 
  truth clusters.
  
  The provided ground truth may or may not cover the entire data set on which 
  prediction is performed. However, we assume that a given ground truth 
  cluster itself is complete (For a given ground truth cluster {a,b,c}, we 
  assume that there could be no other unknown records which should appear
  in it - that is no missed links - and none of the records in {a.b.c} create 
  wrong links - that is no false links). We also assume that all the records
  in the ground truth appear in the prediction as well.
  
  To run this code, the user must provide the ground truth file
  (ground_truth.csv) and prediction file (predicted_clusters.csv) both of
  which are of the following format: 
  file format 
  -----------
    <num_clusers>
    c1_a, c1_b, ...c1_x, <optional comment>
    c2_a, c2_b, ...c2_y, <optional comment>
    .....
    
  It is assumed that these files contain clusters (not record pairs). The
  predicted file contains records grouped by some computer based record
  linkage program. For more information regarding the file format, please
  refer to our documentation in 'https://dmm.anu.edu.au/HISTRL/'.

  Usage: 
    python evaluation_program.py <ground truth> <prediction>
  
  Please note that this program is written for python 2.7.
"""

import csv
import sys
import system_utils

def sim_of_gt_and_pred_clust(gt_clusters_to_process, dict_pred_cluster_of_ind, \
               miss_grp_mem_ind_set, wrong_assn_mem_ind_set):
  '''
  This function attempts to do a similarity calculation between unprocessed
  ground truth clusters and predicted clusters (which are not part of GG_E). 
  
  For each ground truth cluster (GT), we find the clusters in the prediction
  where the nodes from GT appear. Then, a similarity  calculation between the
  GT clusters and the corresponding predicted clusters is done, based on which
  we later do a 1-1 mapping of GT and predicted clusters.
  
  input
  -----
    gt_clusters_to_process - Ground truth clusters which do not have an 
                 exact match (GG_E) in the prediction.
                 
    dict_pred_cluster_of_ind - Dictionary containing an individual 
                   (record ID) as the key and the corresponding 
                   predicted cluster as the value.
                   
    miss_grp_mem_ind_set - Set of record IDs which appear in a group in the
                 ground truth and as a singleton in the 
                 prediction (GS).
                 
    wrong_assn_mem_ind_set - Records which were assigned to the wrong group 
                 in the prediction (GG_W).
  
  output
  ------
    gt_pred_pair_sim_list - A list of tuples, where each tuple contains
                the similarity between a ground truth cluster
                and a predicted cluster as Jaccard similarity,
                true link similarity, GT cluster, predicted
                cluster. Please read our journal paper,
                'Evaluation measure for group-based record 
                linkage' for more information on similarity
                calculation.
                
    new_gt_clusters_to_process - The ground truth clusters which are not 
                   split completely in the prediction, and are
                   therefore retained for further processing.

  '''
  
  # Create a list to store the tuples of GT/Predicted cluster pair 
  # similarities and the corresponding cluster pair
  gt_pred_pair_sim_list = [] 
  
  # Number of completely split GT clusters 
  # (where each node appears in a different cluster)
  complete_split_count = 0
  
  # make a new set of GT clusters to process. This would not contain
  # completely split clusters.
  new_gt_clusters_to_process = set()
  
  # Loop through the ground truth clusters to process and find 
  # similarities with all the predicted clusters which contain  
  # at least 2 nodes from the GT. The similarity is calculated 
  # Jaccard and true link measures.
  for gt_cluster in gt_clusters_to_process:
    set_of_pred_clust = set() # Set of all predicted clusters in which  
              # individuals from the ground truth cluster appear
    
    complete_split = True  # A parameter to monitor the clusters which were 
              # completely split 
    
    for ind in gt_cluster:
      if (ind not in miss_grp_mem_ind_set):
        pred_cluster_of_ind = dict_pred_cluster_of_ind[ind]
        set_of_pred_clust.add(pred_cluster_of_ind)
        
    for pred_clus in set_of_pred_clust:
      common_nodes = set(gt_cluster) & set(pred_clus)
      true_link_sim = len(common_nodes)  # Calculate the true link 
                      # similarity between ground truth 
                      # and predicted cluster.
      
      if(true_link_sim > 1):
        # Calculate the Jaccard similarity only if the true link similarity
        # is greater than one, and populate the gt_pred_pair_sim_list.
        jaccard_sim = \
        float(true_link_sim)/len(set(gt_cluster).union(set(pred_clus)))
        gt_pred_pair_sim_list.append((jaccard_sim, true_link_sim, gt_cluster, 
                                      pred_clus))
       
        complete_split = False  # If at least one associated predicted cluster 
        # has more than 1 node from the GT cluster, then the cluster was not 
        # completely split.
        
      else:
        # Add the nodes which appear in a wrong cluster to 
        # wrong_assn_mem_ind_set
        assert (true_link_sim == 1 and len(pred_clus) > 1)
        wrong_assn_mem_ind_set.add(common_nodes.pop()) 
        
    if (complete_split):
      complete_split_count += 1  # Increment complete split count
      
    else:
      # Only retain the GT clusters which are not completely split for
      # further processing.
      new_gt_clusters_to_process.add(gt_cluster)
      
  print 'The number of ground truth clusters which were ' + \
        'completely split = %d' % (complete_split_count) 
  print
  
  assert (len(new_gt_clusters_to_process) == len(gt_clusters_to_process) - 
                                             complete_split_count)
  
  return gt_pred_pair_sim_list, new_gt_clusters_to_process
  
# -----------------------------------------------------------------------------


def one_to_one_mapping(gt_pred_pair_sim_list):
  '''
  This method executes a one-to-one match based on the similarities 
  calculated between GT and predicted cluster pairs (as indicated
  in gt_pred_pair_sim_list).
  We are using a greedy approach for the 1-1 assignment, where
  each GT cluster is assigned to the best match based on a sorting of the
  similarity value(s).
  
  input
  -----
    gt_pred_pair_sim_list - A list of tuples, where each tuple contains
                the similarity between a ground truth cluster
                and a predicted cluster as Jaccard similarity,
                true link similarity, GT cluster, predicted 
                cluster. Please read our journal paper, 
                'Evaluation measure for group-based record linkage'
                for more information on similarity calculation.
                
  output
  ------
    mapping_dict - A dictionary to store the 1-1 mapping. The key is the best 
             predicted cluster and the value is the ground truth cluster 
             with which it was mapped.
  '''

  mapping_dict = {}  # The output dictionary.
  
  mapped_gt_clusters = set() # Set to keep track of already mapped GT clusters.
  
  best_match_overlap = 0  # Count overlaps for best match.
  
  # Since we follow a greedy method we give highest priority to the  
  # highest similarity, and least priority to the lowest similarity. 
  # Therefore sort gt_pred_pair_sim_list in reverse order.
  gt_pred_pair_sim_list.sort(reverse=True)    
  
  # Iterate through the sorted similarity calculations and assign each 
  # GT cluster to a unique predicted cluster
  for sim_tuple in gt_pred_pair_sim_list:
    
    gt_cluster = sim_tuple[-2]
    
    if (gt_cluster in mapped_gt_clusters):
      continue  # If the GT cluster is already processed, ignore this
    
    best_pred_match = sim_tuple[-1]
    
    # There should be more than one true link between the two clusters
    assert (len(best_pred_match) > 1 and \
        len(set(gt_cluster) & set(best_pred_match)) > 1), \
        (gt_cluster, best_pred_match)
    
    if (best_pred_match not in mapping_dict):
      # If the best cluster was not already assigned to a different 
      # ground truth cluster, map this cluster pair.
      mapping_dict[best_pred_match] = gt_cluster
      mapped_gt_clusters.add(gt_cluster)  # Mark the GT cluster as processed
    
    else:
      best_match_overlap += 1  # Count the times mappings were skipped due
      # to the predicted cluster already being mapped to a GT cluster.
  
  print 'The number of ground truth clusters which had ' + \
  'an overlap for the best match = %d' % (best_match_overlap)
  print
  
  print 'The number of ground truth clusters with a ' + \
  '1-1 mapping (excluding exact matches) = %d' % (len(mapping_dict))
  print
  
  return mapping_dict       

# -----------------------------------------------------------------------------

  
def evaluation_func(gt_file, pred_file):
  '''
  This function executes our proposed evaluation technique on clustering results
  as proposed in our journal paper.
  
  input
  -----
    gt_file -   This CSV file contains the ground truth clusters in the format
          defined in our documentation at 'https://dmm.anu.edu.au/HISTRL/'.
          
    pred_file - This CSV file contains the predicted clusters in the format
          defined in our documentation at 'https://dmm.anu.edu.au/HISTRL/'.
          
  output
  ------
    dict_eval_counts - This dictionary contains the number of records 
               categorized into each of the seven categories defined in our 
               evaluation method (SS, SG, GS, GG_E, GG_M, GG_m, and GG_W).
               The key is the category name and the value is the count. 
          
  '''
  if ('conn_comp_clustering' in pred_file):
    clustering_algo = 'connected component clustering'
  elif ('robust_graph_clustering' in pred_file):
    clustering_algo = 'robust graph clustering'
  else:
    print 'Unknown algorithm!'
    sys.exit()
  
  dict_eval_counts = {} # Dictionary to contain count of each of 7 categories.
  complete_gt = True  # Whether the ground truth covers the entire data set. 
  
  gt_cluster_list = system_utils.load_cluster_file(gt_file)  # Ground truth 
                                                             # cluster list.
  pred_cluster_list = system_utils.load_cluster_file(pred_file) # Predicted 
                                                                # cluster list.
  
  sim_threshold = float(pred_file.replace('.csv','').split('_')[-1])
  
  print 'Running the evaluation function for results generated with the ' + \
        '%s algorithm with similarity threshold %.2f ...' %(clustering_algo, 
                                                        sim_threshold)
  print
    
  gt_record_set = set()
  pred_record_set = set()
  
  for gt_cluster in gt_cluster_list:
    gt_record_set = gt_record_set.union(set(gt_cluster))
    
  for pred_cluster in pred_cluster_list:
    pred_record_set = pred_record_set.union(set(pred_cluster))
    
  assert len(gt_record_set - pred_record_set) == 0, \
  len(gt_record_set - pred_record_set)  # Assert there are no
  # records which only appear in the ground truth clusters.
  
  rec_in_pred_only = pred_record_set - gt_record_set 
  
  if (len(rec_in_pred_only) > 0):
    print 'The ground truth does not cover the complete data set. ' + \
    'There are %d records in the prediction ' % (len(rec_in_pred_only)) + \
    'which are not included in the ground truth.'
    complete_gt = False
    
  else:
    assert gt_record_set == pred_record_set
    print 'The ground truth covers the complete data set. ' + \
    'Records contained in ground truth clusters and predicted ' + \
    'clusters are common to both.'
  print
  
  # Dictionary to store the predicted cluster (value) of a record ID (key)
  dict_pred_cluster_of_ind = {}
  
  gt_sing_set = set()
  gt_cluster_set = set()
  pred_sing_set = set()
  pred_cluster_set = set()
  
  # Create sets to contain the individuals (records) identified as GG_E,
  # majority (GG_M)/minority (GG_m) group match, or 
  # wrongly assigned member (GG_W).
  exact_match_ind_set = set()
  maj_match_ind_set = set()
  min_match_ind_set = set()
  wrong_assn_mem_ind_set = set()
  
  # Total number of nodes in GT clusters (groups)
  gt_cluster_node_count = 0    
  # Total number of nodes in predicted clusters (groups)
  pred_cluster_node_count = 0    
  
  # ---------------------------------------------------------------------------
  # Separate the singletons and the clusters in both the ground truth and 
  # the prediction
  for cluster in gt_cluster_list: 
    if (len(cluster) == 1):
      gt_sing_set.add(cluster[0])
    else:
      gt_cluster_set.add(cluster)
      gt_cluster_node_count += len(cluster)
      
  for cluster in pred_cluster_list: 
    if (len(cluster) == 1):
      pred_sing_set.add(cluster[0])
    else:
      pred_cluster_set.add(cluster)
      pred_cluster_node_count += len(cluster)
      
  print 'Ground truth and predicted cluster/singleton statistics >>> '
  print 'Total number of singletons in the ground truth = %d' % \
  (len(gt_sing_set))
  print 'Total number of singletons in the prediction = %d' % \
  (len(pred_sing_set))
  print 'Total number of ground truth clusters (groups) = %d' % \
  (len(gt_cluster_set))
  print 'Total number of predicted clusters (groups) = %d' % \
  (len(pred_cluster_set))
  print 'Total number of records in groups in the ground truth = %d' % \
  (gt_cluster_node_count)
  print 'Total number of records in groups in the prediction = %d' % \
  (pred_cluster_node_count)
  print
      
  # ---------------------------------------------------------------------------
  # Calculate SS, SG, GS, and GG_E
  
  # Find the correct singletons (SS)
  corr_sing_set = gt_sing_set & pred_sing_set
  
  # Find the wrongly grouped singletons (SG).
  # This would be the singletons in the 
  # ground truth which are not correct singletons
  wrong_grp_sin_set = gt_sing_set - corr_sing_set
  
  # Find the exact matching clusters (GG_E)
  exact_match = gt_cluster_set & pred_cluster_set
  
  # Populate the exact_match_ind_set for GG_E.
  for cluster in exact_match:
    for ind in cluster:
      exact_match_ind_set.add(ind)
  
  # Find the missed group members (GS) - nodes that appear in a
  # cluster in the ground truth, and as a singleton in the prediction
  miss_grp_mem_ind_set = (pred_sing_set - corr_sing_set) & gt_record_set
  
  # ---------------------------------------------------------------------------
  # Find the clusters in the ground truth/ prediction 
  # which are yet to be processed (not exact matches)
  gt_clusters_to_process = gt_cluster_set - exact_match
  pred_clusters_to_process = pred_cluster_set - exact_match
      
  # Populate dict_pred_cluster_of_ind indicating the predicted cluster
  # in which an individual (record ID) appears
  for cluster in pred_clusters_to_process:
    for ind in cluster:
      dict_pred_cluster_of_ind[ind] = cluster
  
  # Execute a similarity calculation for one-to-one mapping of 
  # ground truth clusters and predicted clusters
  gt_pred_pair_sim_list, gt_clusters_to_process = sim_of_gt_and_pred_clust(
                              gt_clusters_to_process,
                              dict_pred_cluster_of_ind,
                              miss_grp_mem_ind_set,
                              wrong_assn_mem_ind_set)
  
  # Execute one-to-one mapping of clusters, based on the similarity calculation.
  mapping_dict = one_to_one_mapping(gt_pred_pair_sim_list)
  
  maj_match_cluster_count = 0  # The GG_M cluster count
  min_match_cluster_count = 0  # The GG_m cluster count
  
  # First process the GT clusters with a 1-1 mapping
  for best_pred_clus, gt_cluster in mapping_dict.iteritems():
    assert (gt_cluster in gt_clusters_to_process)
    
    # Remove cluster from gt_clusters_to_process
    gt_clusters_to_process.remove(gt_cluster)
    
    common_nodes = set(gt_cluster) & set(best_pred_clus)
    
    if (len(common_nodes) / float(len(best_pred_clus)) >= 0.5):
      # If at least 50% of nodes in the best prediction come from the GT,
      # cluster, we call it a majority cluster (GG_M). Therefore, all 
      # nodes common to the best predicted cluster and ground truth 
      # cluster are assigned to the GG_M category.
      maj_match_ind_set = maj_match_ind_set.union(common_nodes)
      maj_match_cluster_count += 1
      
    else:
      assert (len(common_nodes) / float(len(best_pred_clus)) < 0.5)
      # If less than 50% of nodes in the best prediction come from the GT,
      # cluster, we call it a minority cluster (GG_m). Therefore, all 
      # nodes common to the best predicted cluster and ground truth 
      # cluster are assigned to the GG_m category.
      min_match_ind_set = min_match_ind_set.union(common_nodes)
      min_match_cluster_count += 1
     
    nodes_to_process = set(gt_cluster) - common_nodes  # Nodes left to be 
    # processed. i.e. nodes from the ground truth cluster which were not 
    # classified as GG_M or GG_m.
    
    for node in nodes_to_process:
      if (node not in miss_grp_mem_ind_set):
        wrong_assn_mem_ind_set.add(node)  # Nodes which were not identified 
                       # as GS should be categorized as GG_W.
  
  for gt_cluster in gt_clusters_to_process:
    # Process the ground truth clusters which do not have a mapping
    for node in gt_cluster:
      if (node not in miss_grp_mem_ind_set):
        wrong_assn_mem_ind_set.add(node)
        
  # Verify results with assertions
  assert (len(gt_sing_set) == len(corr_sing_set) + len(wrong_grp_sin_set))
  assert (gt_cluster_node_count == len(miss_grp_mem_ind_set) + \
      len(exact_match_ind_set) + len(maj_match_ind_set) + \
      len(min_match_ind_set) + len(wrong_assn_mem_ind_set))
  
  if (complete_gt):
    assert (len(pred_sing_set) == len(corr_sing_set) + \
            len(miss_grp_mem_ind_set))
    assert (pred_cluster_node_count == len(wrong_grp_sin_set) + \
        len(exact_match_ind_set) + len(maj_match_ind_set) + \
        len(min_match_ind_set) + len(wrong_assn_mem_ind_set))
  
  total_in_categories = len(corr_sing_set) + len(miss_grp_mem_ind_set) + \
      len(wrong_grp_sin_set) + len(exact_match_ind_set) + \
      len(maj_match_ind_set) + len(min_match_ind_set) + \
      len(wrong_assn_mem_ind_set)
  assert (total_in_categories == len(gt_record_set)), (total_in_categories, \
                                                       len(gt_record_set)) 
  
  # print the number of ground truth clusters categorized as 
  # GG_E, GG_M, and GG_m.
  print 'Number of ground truth clusters identified to have a mapping to a ' + \
      'predicted cluster as GG_E, GG_M, or GG_m >>>'
  print 'Total number of exact matching clusters = %d' % (len(exact_match))
  print 'Total number of majority matching clusters = %d' % \
  (maj_match_cluster_count)
  print 'Total number of minority matching clusters = %d' % \
  (min_match_cluster_count)
  print
      
  # print the number of individuals in each category.
  print '----------------------------------------------------------------------'
  print 'Number of records assigned to each of 7 categories'
  print 'Correct singleton (SS)         = %d' % (len(corr_sing_set))
  print 'Wrongly grouped singleton (SG) = %d' % (len(wrong_grp_sin_set))
  print 'Missed group member (GS)       = %d' % (len(miss_grp_mem_ind_set))
  print 'Exact group match (GG_E)       = %d' % (len(exact_match_ind_set))
  print 'Majority group match (GG_M)    = %d' % (len(maj_match_ind_set))
  print 'Minority group match (GG_m)    = %d' % (len(min_match_ind_set))
  print 'Wrongly assigned member (GG_W) = %d' % (len(wrong_assn_mem_ind_set))
  print '----------------------------------------------------------------------'
  print
  
# -----------------------------------------------------------------------------

# Call the evaluation_func method for execution.
if __name__== "__main__":
    assert len(sys.argv) == 3, len(sys.argv)
    evaluation_func(sys.argv[1], sys.argv[2])
    
# -----------------------------------------------------------------------------
    